/*** prostat 1.32 ***/

#include "analhead2.h"

/*** hash.c; the functions which do all the work in the hash tables. ***/

/*** The first few are all exactly the same, adding a certain number of
  requests and a certain number of bytes to the hash entry for that item,
  creating a new hash entry if none existed before. ***/

extern void *xmalloc();    /* in utilities.c */
                           /* used in so many functions, I declare it here */

void reqhashadd(char *sitename, int reqs, double bytes, int reqs_hit, flag al)
{
  extern int magicno();             /* in utilities.c */

  extern struct url *urlhead[];
  extern int no_urls;

  int magicnumber;
  flag finished;
  struct url *urlp, *urllastp, *urlprevp;

  /* First calculate sitename's "magic number" */

  magicnumber = magicno(sitename, URLHASHSIZE);

  /* Now look through the magicnumber'th list for that URL */

  finished = FALSE;
  urlp = (urlhead[magicnumber]);
  urllastp = urlp;
  urlprevp = urlp;
  while (urlp -> name != NULL && !finished) {
    if (strcmp(urlp -> name, sitename) == 0) {   /* then done */
      urlp -> reqs += reqs;
      urlp -> bytes += bytes;
      if(reqs_hit == 1) {
        urlp -> reqs_hit += 1;
        urlp -> bytes_hit += bytes;
      }
      /* keep the list in rough (not exact) order of number of
	 accesses for quicker searching; particularly useful if
	 the HASHSIZE is set too low. */
      if (urlp -> reqs > urllastp -> reqs) {
	if (urlprevp == urllastp) {   /* iff urlp is 2nd in the list */
	  urlhead[magicnumber] = urlp;
	  urllastp -> next = urlp -> next;
	  urlp -> next = urllastp;
	}
	else {  /* urlp is 3rd or later in the list */
	  urlprevp -> next = urlp;
	  urllastp -> next = urlp -> next;
	  urlp -> next = urllastp;
	}
      }
      finished = TRUE;
    }
    else {      /* look at the next one */
      urlprevp = urllastp;
      urllastp = urlp;
      urlp = urlp -> next;
    }
  }
  
  if (!finished) {   /* reached the end of the list without success; new URL */
    no_urls++;
    urlp -> name = (char *) xmalloc((size_t)(strlen(sitename) + 1));
    strcpy(urlp -> name, sitename);
    urlp -> reqs = reqs;
    urlp -> bytes = bytes;
    urlp -> bytes = bytes;
    if(reqs_hit == 1) {
        urlp -> reqs_hit = 1;
        urlp -> bytes_hit = bytes;
    }
    else {
        urlp -> reqs_hit = 0;
        urlp -> bytes_hit = 0;
    }
    urlp -> aliasdone = al;
    urlp -> next = (struct url *) xmalloc(sizeof(struct url));
    urlp -> next -> name = NULL;
  }

}

void requesthashadd(char *requestname, int reqs, double bytes, int time)
{
  extern int magicno();          /* in utilities.c */
  extern int debug;

  extern struct request *requesthead[];

  int magicnumber;
  flag finished;
  struct request *requestp, *requestlastp, *requestprevp;

  magicnumber = magicno(requestname, PROTOHASHSIZE);

  finished = FALSE;
  requestp = (requesthead[magicnumber]);
  requestlastp = requestp;
  requestprevp = requestp;
  while (requestp -> name != NULL && !finished) {
    if (strcmp(requestp -> name, requestname) == 0) {
      requestp -> reqs += reqs;
      requestp -> bytes += bytes;
      requestp -> time += time;
      if (requestp -> reqs > requestlastp -> reqs) {
        if (requestprevp == requestlastp) {
          requesthead[magicnumber] = requestp;
          requestlastp -> next = requestp -> next;
          requestp -> next = requestlastp;
        }
        else {
          requestprevp -> next = requestp;
          requestlastp -> next = requestp -> next;
          requestp -> next = requestlastp;
        }
      }
      finished = TRUE;
    }
    else {
      requestprevp = requestlastp;
      requestlastp = requestp;
      requestp = requestp -> next;
    }
  }

  if (!finished) {
    requestp -> name = (char *) xmalloc((size_t)(strlen(requestname) + 1));
    strcpy(requestp -> name, requestname);
    requestp -> reqs = reqs;
    requestp -> bytes = bytes;
    requestp -> time = time;
    requestp -> next = (struct request *) xmalloc(sizeof(struct request));
    requestp -> next -> name = NULL;
  }

}

void protohashadd(char *protoname, int reqs, double bytes)
{
  extern int magicno();          /* in utilities.c */

  extern struct proto *protohead[];

  int magicnumber;
  flag finished;
  struct proto *protop, *protolastp, *protoprevp;

  magicnumber = magicno(protoname, PROTOHASHSIZE);

  finished = FALSE;
  protop = (protohead[magicnumber]);
  protolastp = protop;
  protoprevp = protop;
  while (protop -> name != NULL && !finished) {
    if (strcmp(protop -> name, protoname) == 0) {
      protop -> reqs += reqs;
      protop -> bytes += bytes;
      if (protop -> reqs > protolastp -> reqs) {
	if (protoprevp == protolastp) {
	  protohead[magicnumber] = protop;
	  protolastp -> next = protop -> next;
	  protop -> next = protolastp;
	}
	else {
	  protoprevp -> next = protop;
	  protolastp -> next = protop -> next;
	  protop -> next = protolastp;
	}
      }
      finished = TRUE;
    }
    else {
      protoprevp = protolastp;
      protolastp = protop;
      protop = protop -> next;
    }
  }
  
  if (!finished) {
    protop -> name = (char *) xmalloc((size_t)(strlen(protoname) + 1));
    strcpy(protop -> name, protoname);
    protop -> reqs = reqs;
    protop -> bytes = bytes;
    protop -> next = (struct proto *) xmalloc(sizeof(struct proto));
    protop -> next -> name = NULL;
  }

}


void hosthashadd(char *hostn, int reqs, double bytes, int time, int reqs_hit, int tcp_req, flag is_an_alias)
{
  extern int magicno();          /* in utilities.c */

  extern struct host *hosthead[];
  extern int no_hosts;

  int magicnumber;
  flag finished;
  struct host *hostp, *hostlastp, *hostprevp;

  magicnumber = magicno(hostn, HOSTHASHSIZE);

  finished = FALSE;
  hostp = (hosthead[magicnumber]);
  hostlastp = hostp;
  hostprevp = hostp;

  while (hostp -> name != NULL && !finished) {
    if (strcmp(hostp -> name, hostn) == 0) {
      hostp -> reqs += reqs;
      hostp -> bytes += bytes;
      hostp -> time += time;
      if(reqs_hit == 1) {
      	hostp -> reqs_hit += 1;
      	hostp -> bytes_hit += bytes;
      }

      if(tcp_req == 1) {
      	hostp -> reqs_tcp += 1;
	hostp -> bytes_tcp += bytes;
      }
      else {
      	hostp -> reqs_udp += 1;
	hostp -> bytes_udp += bytes;
      }

      if (hostp -> reqs > hostlastp -> reqs) {
	if (hostprevp == hostlastp) {
	  hosthead[magicnumber] = hostp;
	  hostlastp -> next = hostp -> next;
	  hostp -> next = hostlastp;
	}
	else {
	  hostprevp -> next = hostp;
	  hostlastp -> next = hostp -> next;
	  hostp -> next = hostlastp;
	}
      }
      finished = TRUE;
    }
    else {
      hostprevp = hostlastp;
      hostlastp = hostp;
      hostp = hostp -> next;
    }
  }
  
  if (!finished) {
    no_hosts++;
    hostp -> name = xmalloc((size_t)(strlen(hostn) + 1));
    strcpy(hostp -> name, hostn);
    hostp -> is_an_alias = is_an_alias;
    hostp -> reqs = reqs;
    hostp -> bytes = bytes;
    hostp -> time = time;
    if(reqs_hit == 1) {
    	hostp -> reqs_hit = 1;
    	hostp -> bytes_hit = bytes;
    }
    else {
    	hostp -> reqs_hit = 0;
    	hostp -> bytes_hit = 0;
    }
    if(tcp_req == 1) {
      	hostp -> reqs_tcp += 1;
	hostp -> bytes_tcp += bytes;
	hostp -> reqs_udp = 0;
    	hostp -> bytes_udp = 0;
    }
    else {
    	hostp -> reqs_tcp = 0;
    	hostp -> bytes_tcp = 0;
	hostp -> reqs_udp += 1;
	hostp -> bytes_udp += bytes;
    }
    
    hostp -> is_a_neighbour = OFF;
    hostp -> r_reqs = 0;
    hostp -> r_bytes = 0;
    hostp -> next = (struct host *) xmalloc(sizeof(struct host));
    hostp -> next -> name = NULL;
  }
}

void neighbourhashadd(char *neighbour, int reqs, double bytes)
{
  extern int magicno();          /* in utilities.c */

  extern struct host *hosthead[];
  extern int no_hosts;
  extern flag neighbours;
  
  struct hostent *host;
  struct in_addr ptr;
  int magicnumber;
  flag finished;
  struct host *hostp, *hostlastp, *hostprevp;
  char hostn[MAXSTRINGLENGTH];
  unsigned long addr;


  neighbours = ON;
  addr = inet_addr(neighbour);
  if (addr == INADDR_NONE) {
  	host = gethostbyname(neighbour);
	ptr = *(struct in_addr *)host->h_addr;
	sprintf(hostn, "%s", inet_ntoa(ptr));
  }
  else
  	strcpy(hostn, neighbour);
  	

  magicnumber = magicno(hostn, HOSTHASHSIZE);

  finished = FALSE;
  hostp = (hosthead[magicnumber]);
  hostlastp = hostp;
  hostprevp = hostp;
  while (hostp -> name != NULL && !finished) {
    if (strcmp(hostp -> name, hostn) == 0) {
      hostp -> is_a_neighbour = ON;
      hostp -> r_reqs += reqs;
      hostp -> r_bytes += bytes;
      finished = TRUE;
    }
    else {
      hostprevp = hostlastp;
      hostlastp = hostp;
      hostp = hostp -> next;
    }
  }
  
  if (!finished) {
    no_hosts++;
    hostp -> name = xmalloc((size_t)(strlen(hostn) + 1));
    strcpy(hostp -> name, hostn);
    hostp -> reqs = 0;
    hostp -> bytes = 0;
    hostp -> reqs_hit = 0;
    hostp -> is_a_neighbour = ON;
    hostp -> r_reqs = reqs;
    hostp -> r_bytes = bytes;
    hostp -> next = (struct host *) xmalloc(sizeof(struct host));
    hostp -> next -> name = NULL;
  }
  
  
}

/*** The domain hashadd function is different from the other three, because we
   already know all domains, so there need be no clashes in the hash table. ***/

void domhashadd(char *hostn, int reqs, double bytes, int reqs_hit)
{
  extern int hosttodomcode();           /* in alias.c */
  extern int no_doms;

  extern struct domain *domainhead[];
  extern int debug;

  int domcode;   /* domcode is as explained at the bit where we read the
		    domains in. It's the equivalent of magic number for
		    the other hashadd functions. */

  domcode = hosttodomcode(hostn);
  if (domcode == DOMHASHSIZE - 2 && debug >= 2)
    fprintf(stderr, "U: %s\n", hostn);

  if(domainhead[domcode] -> reqs == 0) no_doms++;
  domainhead[domcode] -> reqs += reqs;
  domainhead[domcode] -> bytes += bytes;
  if(reqs_hit == 1) {
        domainhead[domcode] -> reqs_hit += 1;
        domainhead[domcode] -> bytes_hit += bytes;
  }

}

void mimehashadd(char *type, int reqs, double bytes)
{
 
  extern int magicno();          /* in utilities.c */

  extern struct mime_types *mimehead[];
  extern int no_mimes;

  int magicnumber;
  flag finished;
  struct mime_types *mimep, *mimelastp, *mimeprevp;

  magicnumber = magicno(type, TYPEHASHSIZE);

  finished = FALSE;
  mimep = (mimehead[magicnumber]);
  mimelastp = mimep;
  mimeprevp = mimep;
  while (mimep -> name != NULL && !finished) {
    if (strcmp(mimep -> name, type) == 0) {
      mimep -> reqs += reqs;
      mimep -> bytes += bytes;
      if (mimep -> reqs > mimelastp -> reqs) {
        if (mimeprevp == mimelastp) {
          mimehead[magicnumber] = mimep;
          mimelastp -> next = mimep -> next;
          mimep -> next = mimelastp;
        }
        else {
          mimeprevp -> next = mimep;
          mimelastp -> next = mimep -> next;
          mimep -> next = mimelastp;
        }
      }
      finished = TRUE;
    }
    else {
      mimeprevp = mimelastp;
      mimelastp = mimep;
      mimep = mimep -> next;
    }
  }

  if (!finished) {
    no_mimes++;
    mimep -> name = xmalloc((size_t)(strlen(type) + 1));
    strcpy(mimep -> name, type);
    mimep -> reqs = reqs;
    mimep -> bytes = bytes;
    mimep -> next = (struct mime_types *) xmalloc(sizeof(struct mime_types));
    mimep -> next -> name = NULL;
  }


}


void typehashadd(char *type, char *typename, int reqs, double bytes)
{
  
  extern int magicno();          /* in utilities.c */
  extern struct file_types *typehead[];
  extern int no_types;
  extern flag mq;

  int magicnumber;
  flag finished;
  struct file_types *typep, *typelastp, *typeprevp;

  magicnumber = magicno(type, TYPEHASHSIZE);

  finished = FALSE;
  typep = (typehead[magicnumber]);
  typelastp = typep;
  typeprevp = typep;
  while (typep -> name != NULL && !finished) {
    if (strcmp(typep -> name, type) == 0) {
      if(mq) mimehashadd(typep -> real_name, reqs, bytes);
      typep -> reqs += reqs;
      typep -> bytes += bytes;
      if (typep -> reqs > typelastp -> reqs) {
        if (typeprevp == typelastp) {
          typehead[magicnumber] = typep;
          typelastp -> next = typep -> next;
          typep -> next = typelastp;
        }
        else {
          typeprevp -> next = typep;
          typelastp -> next = typep -> next;
          typep -> next = typelastp;
        }
      }
      finished = TRUE;
    }
    else {
      typeprevp = typelastp;
      typelastp = typep;
      typep = typep -> next;
    }
  }


  if (!finished && reqs == 0) { 
    no_types++;
    typep -> name = xmalloc((size_t)(strlen(type) + 1));
    strcpy(typep -> name, type);
    typep -> real_name = xmalloc((size_t)(strlen(typename) + 1));
    strcpy(typep -> real_name, typename);
    typep -> reqs = reqs;
    typep -> bytes = bytes;
    typep -> next = (struct file_types *) xmalloc(sizeof(struct file_types));
    typep -> next -> name = NULL;
  }

}


/*** add 1 to the number of requests for a particular month ***/
void addmonthlydata(int year, int monthno)
{
  extern struct monthly *firstm;
  extern struct timestruct firsttime;

  struct monthly *mp;
  int i;

  mp = firstm;

  for (i = year - firsttime.year; i > 0; i--)
    mp = mp -> next;       /* run on to the right year */

  mp -> reqs[monthno]++;   /* and add 1 to the right month in that year */

}

/* add 1 to the number of requests for a particular day */
void adddailydata(int year, int monthno, int date)
{
  extern struct daily *firstd;
  extern struct timestruct firsttime;

  struct daily *dp;
  int i;

  dp = firstd;

  for (i = (year - firsttime.year) * 12 + monthno - firsttime.monthno;
       i > 0; i--)
    dp = dp -> next;       /* run on to the right month */

  dp -> reqs[date - 1]++;   /* and add 1 to the right date in that month */

}

/* add 1 to the number of requests for a particular week */
void addweeklydata(int year, int monthno, int date)
{
  extern int minsbetween();     /* in utilities.c */

  extern struct weekly *firstw;

  struct weekly *wp;

  wp = firstw;

  while(minsbetween(wp -> start.date, wp -> start.monthno, wp -> start.year,
		    0, 0, date, monthno, year, 0, 0) >= 10080)
    wp = wp -> next;       /* run on to the right week */

  wp -> reqs++;            /* and add 1 to it */

}

/*** Now some routines to sort the various reports ready for printing ***/

struct url *reqsort(void)
{   /* sort the request report */

  extern flag ispage();           /* in alias.c */

  extern int rnumber;
  extern struct url *urlhead[];
  extern int reqtype;
  extern int sitessortby;
  extern int sitesfloor;
  extern double total_bytes;
  extern int url_max_reqs;
  extern double url_max_bytes;
  extern int no_reqsort;

  flag wantit = ON;    /* whether we want a particular URL */
  struct url *urlsorthead;   /* build up the sort in this list
				(and return it at the end) */
  struct url *urlp, *urlp2, *urlnextp, *urllastp;
  int onlist;          /* the list we are processing */
  flag finished;       /* whether we've finished with a particular entry */

  int i;

  rnumber = 0;
  urlsorthead = (struct url *) xmalloc(sizeof(struct url));
  urlsorthead -> name = NULL;        /* as marker */
  onlist = 0;                        /* the list we are on */
  urlp = urlhead[0];                 /* starting at list 0 */
  for ( ; onlist < URLHASHSIZE; urlp = urlnextp)  {
                                     /* run through all the URLs */
    if (urlp -> name == NULL)    {   /* then finished this list */
      urlnextp = urlhead[++onlist];  /* so look at the next list */
    }
    else {
      if (reqtype == PAGES)
	wantit = ispage(urlp -> name);  /* o/wise wantit is always ON */
      if ((sitessortby == BYBYTES &&
	   (urlp -> bytes / (total_bytes / 10000)) < sitesfloor) ||
	  (sitessortby != BYBYTES && urlp -> reqs < sitesfloor) ||
	  (!wantit)) {  /* we don't want it */
	urlnextp = urlp -> next;
      }
      else {
	no_reqsort++;
	rnumber++;
	url_max_reqs = MAX(urlp -> reqs, url_max_reqs);
	url_max_bytes = MAX(urlp -> bytes, url_max_bytes);
	if ((urlsorthead -> name == NULL) ||
	    (sitessortby == BYBYTES && urlp -> bytes > urlsorthead -> bytes) ||
	    (sitessortby == BYREQUESTS && urlp -> reqs > urlsorthead -> reqs) ){
	  /* if it's before the first item currently on the list, slot it in */
	  urlnextp = urlp -> next;   /* the next one we're going to look at */
	  urlp -> next = urlsorthead;
	  urlsorthead = urlp;
	}
	else {                   /* otherwise compare with the ones so far */
	  finished = OFF;
	  urllastp = urlsorthead;
	  if (sitesfloor < 0)
	    i = sitesfloor;
	  else
	    i = 1;
	  for (urlp2 = urlsorthead -> next;
	       urlp2 -> name != NULL && (!finished) && (i++) != -1;
	       urlp2 = urlp2 -> next) {
	    if ((sitessortby == BYBYTES && urlp -> bytes > urlp2 -> bytes) ||
		(sitessortby == BYREQUESTS && urlp -> reqs > urlp2 -> reqs) ){
	  /* if urlp comes before urlp2 in the chosen ordering, slot it in */
	      urlnextp = urlp -> next;
	      urlp -> next = urlp2;
	      urllastp -> next = urlp;
	      finished = ON;
	    }
	    urllastp = urlp2;
	  }
	  if (!finished) {         /* we've reached the end of the list; */
	                           /* slot it in at the end */
	    urlnextp = urlp -> next;
	    urlp -> next = urlp2;
	    urlp2 -> name = NULL;
	    urllastp -> next = urlp;
	  }
	}
      }
    }
    
    urlp = urlnextp;   /* so, on to the next one */

  }        /* end for running through all filenames */

  return(urlsorthead);

}    /* end reqsort */


struct request *requestsort(void)
{   /* sort the requestcol report */

  extern int inumber;
  extern struct request *requesthead[];
  extern int request_max_reqs;
  extern double request_max_bytes;
  extern int num_type_req_tcp, num_type_req_udp, num_type_req_err;

  struct request *requestsorthead;   /* build up the sort in this list
                                (and return it at the end) */
  struct request *requestp, *requestp2, *requestnextp, *requestlastp;
  int onlist;          /* the list we are processing */
  flag finished;       /* whether we've finished with a particular entry */

  int i;

  inumber = 0;
  requestsorthead = (struct request *) xmalloc(sizeof(struct request));
  requestsorthead -> name = NULL;        /* as marker */
  onlist = 0;                        /* the list we are on */
  requestp = requesthead[0];                 /* starting at list 0 */
  for ( ; onlist < PROTOHASHSIZE; requestp = requestnextp)  {
                                     /* run through all the requests */
    if(requestp -> name != NULL) {
    	if(strstr(requestp -> name, "TCP_") != NULL)
		num_type_req_tcp++;
    	if(strstr(requestp -> name, "UDP_") != NULL)
		num_type_req_udp++;
	if(strstr(requestp -> name, "ERR_") != NULL)
		num_type_req_err++;
    }

    if (requestp -> name == NULL) {      /* then finished this list */
      requestnextp = requesthead[++onlist];  /* so look at the next list */
    }
    else {
      inumber++;
      request_max_reqs = MAX(requestp -> reqs, request_max_reqs);
      request_max_bytes = MAX(requestp -> bytes, request_max_bytes);
      if (requestsorthead -> name == NULL || strcmp(requestp -> name, requestsorthead -> name) < 0)  {
        /* if it's before the first item currently on the list, slot it in */
        requestnextp = requestp -> next;   /* the next one we're going to look at */
        requestp -> next = requestsorthead;
        requestsorthead = requestp;
      }
      else {              /* otherwise compare with the ones so far */
        finished = OFF;
        requestlastp = requestsorthead;
        i = 1;
        for (requestp2 = requestsorthead -> next; requestp2 -> name != NULL
             && (!finished) && (i++) != -1;
             requestp2 = requestp2 -> next) {
	  if (strcmp(requestp -> name, requestp2 -> name) < 0) {
	      requestnextp = requestp -> next;
              requestp -> next = requestp2;
              requestlastp -> next = requestp;
              finished = ON;
	  }
          requestlastp = requestp2;
        }
        if (!finished) {         /* we've reached the end of the list; */
                                 /* slot it in at the end */
          requestnextp = requestp -> next;
          requestp -> next = requestp2;
          requestp2 -> name = NULL;
          requestlastp -> next = requestp;
        }
      }
    }

    requestp = requestnextp;   /* so, on to the next one */

  }        /* end for running through all filenames */

  return(requestsorthead);

}    /* end requestsort */


struct proto *protosort(void)
{   /* sort the protocol report */

  extern int inumber;
  extern struct proto *protohead[];
  extern int protosortby;
  extern int protofloor;
  extern double total_bytes;
  extern int proto_max_reqs;
  extern double proto_max_bytes;

  struct proto *protosorthead;   /* build up the sort in this list
				(and return it at the end) */
  struct proto *protop, *protop2, *protonextp, *protolastp;
  int onlist;          /* the list we are processing */
  flag finished;       /* whether we've finished with a particular entry */

  int i;

  inumber = 0;
  protosorthead = (struct proto *) xmalloc(sizeof(struct proto));
  protosorthead -> name = NULL;        /* as marker */
  onlist = 0;                        /* the list we are on */
  protop = protohead[0];                 /* starting at list 0 */
  for ( ; onlist < PROTOHASHSIZE; protop = protonextp)  {
                                     /* run through all the protos */
    if (protop -> name == NULL) {      /* then finished this list */
      protonextp = protohead[++onlist];  /* so look at the next list */
    }
    else {
      if ((protosortby == BYBYTES &&
	   (protop -> bytes / (total_bytes / 10000)) < protofloor) ||
	  (protosortby != BYBYTES && protop -> reqs < protofloor)) {
                                              /* we don't want it */
	protonextp = protop -> next;
      }
      else {
	inumber++;
	proto_max_reqs = MAX(protop -> reqs, proto_max_reqs);
	proto_max_bytes = MAX(protop -> bytes, proto_max_bytes);
	if ((protosorthead -> name == NULL) ||
	    (protosortby == BYBYTES && protop -> bytes > protosorthead -> bytes) ||
	    (protosortby == BYREQUESTS && protop -> reqs > protosorthead -> reqs) ){
	  /* if it's before the first item currently on the list, slot it in */
	  protonextp = protop -> next;   /* the next one we're going to look at */
	  protop -> next = protosorthead;
	  protosorthead = protop;
	}
	else {              /* otherwise compare with the ones so far */
	  finished = OFF;
	  protolastp = protosorthead;
	  if (protofloor < 0)
	    i = protofloor;
	  else
	    i = 1;
	  for (protop2 = protosorthead -> next; protop2 -> name != NULL
	       && (!finished) && (i++) != -1;
	       protop2 = protop2 -> next) {
	    if ((protosortby == BYBYTES && protop -> bytes > protop2 -> bytes) ||
		(protosortby == BYREQUESTS && protop -> reqs > protop2 -> reqs) ){
	  /* if protop comes before protop2 in the chosen ordering, slot it in */
	      protonextp = protop -> next;
	      protop -> next = protop2;
	      protolastp -> next = protop;
	      finished = ON;
	    }
	    protolastp = protop2;
	  }
	  if (!finished) {         /* we've reached the end of the list; */
	                           /* slot it in at the end */
	    protonextp = protop -> next;
	    protop -> next = protop2;
	    protop2 -> name = NULL;
	    protolastp -> next = protop;
	  }
	}
      }
    }
    
    protop = protonextp;   /* so, on to the next one */

  }        /* end for running through all filenames */

  return(protosorthead);

}    /* end protosort */


struct host *hostsort(void)
{   /* sort the host report */

  extern char *reversehostname();    /* in alias.c */
  extern int hoststrcmp();           /* in utilities.c */

  extern int Snumber;
  extern struct host *hosthead[];
  extern int hostsortby;
  extern int hostfloor;
  extern double total_bytes;
  extern int host_max_reqs;
  extern double host_max_bytes;
  extern int no_hostsort;

  struct host *hostsorthead;   /* build up the sort in this list
				(and return it at the end) */
  struct host *hostp, *hostp2, *hostnextp, *hostlastp;
  int onlist;          /* the list we are processing */
  flag finished;       /* whether we've finished with a particular entry */

  int i;

  Snumber = 0;
  no_hostsort = 0;
  hostsorthead = (struct host *) xmalloc(sizeof(struct host));
  hostsorthead -> name = NULL;        /* as marker */
  onlist = 0;                        /* the list we are on */
  hostp = hosthead[0];                 /* starting at list 0 */
  for ( ; onlist < HOSTHASHSIZE; hostp = hostnextp)  {
                                     /* run through all the HOSTs */
    if (hostp -> name == NULL) {      /* then finished this list */
      hostnextp = hosthead[++onlist];  /* so look at the next list */
    }
    else {
      if ((hostsortby == BYBYTES &&
	   (hostp -> bytes / (total_bytes / 10000)) < hostfloor) ||
	  (hostsortby != BYBYTES && hostp -> reqs < hostfloor)) {
                                              /* we don't want it */
	hostnextp = hostp -> next;
      }
      else {
	Snumber++;
no_hostsort++;
	host_max_reqs = MAX(hostp -> reqs, host_max_reqs);
	host_max_bytes = MAX(hostp -> bytes, host_max_bytes);
	if ((hostsorthead -> name == NULL) ||
	    (hostsortby == BYBYTES && hostp -> bytes > hostsorthead -> bytes) ||
	    (hostsortby == BYREQUESTS &&
	     hostp -> reqs > hostsorthead -> reqs) ){
	  /* if it's before the first item currently on the list, slot it in */
	  hostnextp = hostp -> next;   /* the next one we're going to look at */
	  hostp -> next = hostsorthead;
	  hostsorthead = hostp;
	}
	else {              /* otherwise compare with the ones so far */
	  finished = OFF;
	  hostlastp = hostsorthead;
	  if (hostfloor < 0)
	    i = hostfloor;
	  else
	    i = 1;
	  for (hostp2 = hostsorthead -> next; hostp2 -> name != NULL
	       && (!finished) && (i++) != -1;
	       hostp2 = hostp2 -> next) {
	    if ((hostsortby == BYBYTES && hostp -> bytes > hostp2 -> bytes) ||
		(hostsortby == BYREQUESTS && hostp -> reqs > hostp2 -> reqs) ){
	  /* if hostp comes before hostp2 in the chosen ordering, slot it in */
	      hostnextp = hostp -> next;
	      hostp -> next = hostp2;
	      hostlastp -> next = hostp;
	      finished = ON;
	    }
	    hostlastp = hostp2;
	  }
	  if (!finished) {         /* we've reached the end of the list; */
	                           /* slot it in at the end */
	    hostnextp = hostp -> next;
	    hostp -> next = hostp2;
	    hostp2 -> name = NULL;
	    hostlastp -> next = hostp;
	  }
	}
      }
    }
    
    hostp = hostnextp;   /* so, on to the next one */

  }        /* end for running through all filenames */
/*
  for(hostp = hostsorthead; hostp -> name != NULL; hostp = hostp -> next) {
	no_hostsort++;
  }
*/

  return(hostsorthead);

}    /* end hostsort */

struct mime_types *mimesort(void)
{   /* sort the mime report */


  extern int mnumber;
  extern struct mime_types *mimehead[];
  extern int mimesortby;
  extern double total_bytes;
  extern int mime_max_reqs;
  extern int mimefloor;
  extern double mime_max_bytes;
  extern int no_mimesort;

  struct mime_types *mimesorthead;   /* build up the sort in this list
                                (and return it at the end) */
  struct mime_types *mimep, *mimep2, *mimenextp, *mimelastp;
  int onlist;          /* the list we are processing */
  flag finished;       /* whether we've finished with a particular entry */

  mnumber = 0;
  no_mimesort = 0;
  mimesorthead = (struct mime_types *) xmalloc(sizeof(struct mime_types));
  mimesorthead -> name = NULL;        /* as marker */
  onlist = 0;                        /* the list we are on */
  mimep = mimehead[0];                 /* starting at list 0 */
  for ( ; onlist < TYPEHASHSIZE; mimep = mimenextp)  {
                                     /* run through all the HOSTs */
    if (mimep -> name == NULL) {      /* then finished this list */
      mimenextp = mimehead[++onlist];  /* so look at the next list */
    }
    else {
      if ((mimesortby == BYBYTES &&
           (mimep -> bytes / (total_bytes / 10000)) < mimefloor) ||
          (mimesortby == BYREQUESTS && mimep -> reqs < mimefloor)) {
        mimenextp = mimep -> next;
      }
        else {
        mnumber++;
        no_mimesort++;
        mime_max_reqs = MAX(mimep -> reqs, mime_max_reqs);
        mime_max_bytes = MAX(mimep -> bytes, mime_max_bytes);
        if ((mimesorthead -> name == NULL) ||
            (mimesortby == BYBYTES && mimep -> bytes > mimesorthead -> bytes) ||
            (mimesortby == BYREQUESTS &&
             mimep -> reqs > mimesorthead -> reqs) ) {
          /* if it's before the first item currently on the list, slot it in */
          mimenextp = mimep -> next;   /* the next one we're going to look at */
          mimep -> next = mimesorthead;
          mimesorthead = mimep;
        }
        else {              /* otherwise compare with the ones so far */
          finished = OFF;
          mimelastp = mimesorthead;
          for (mimep2 = mimesorthead -> next; mimep2 -> name != NULL
               && (!finished) ;
               mimep2 = mimep2 -> next) {
            if ((mimesortby == BYBYTES && mimep -> bytes > mimep2 -> bytes) ||
                (mimesortby == BYREQUESTS && mimep -> reqs > mimep2 -> reqs) ) {
          /* if mimep comes before mimep2 in the chosen ordering, slot it in */
              mimenextp = mimep -> next;
              mimep -> next = mimep2;
              mimelastp -> next = mimep;
              finished = ON;
            }
            mimelastp = mimep2;
          }
          if (!finished) {         /* we've reached the end of the list; */
                                   /* slot it in at the end */
            mimenextp = mimep -> next;
            mimep -> next = mimep2;
            mimep2 -> name = NULL;
            mimelastp -> next = mimep;
          }
        }
        }
    }

    mimep = mimenextp;   /* so, on to the next one */

  }        /* end for running through all filenames */

  return(mimesorthead);

}    /* end mimesort */

struct file_types *typesort(void)
{   /* sort the type report */


  extern int tnumber;
  extern struct file_types *typehead[];
  extern int typesortby;
  extern double total_bytes;
  extern int type_max_reqs;
  extern int typefloor;
  extern double type_max_bytes;
  extern int no_typesort;

  struct file_types *typesorthead;   /* build up the sort in this list
                                (and return it at the end) */
  struct file_types *typep, *typep2, *typenextp, *typelastp;
  int onlist;          /* the list we are processing */
  flag finished;       /* whether we've finished with a particular entry */

  tnumber = 0;
  no_typesort = 0;
  typesorthead = (struct file_types *) xmalloc(sizeof(struct file_types));
  typesorthead -> name = NULL;        /* as marker */
  onlist = 0;                        /* the list we are on */
  typep = typehead[0];                 /* starting at list 0 */
  for ( ; onlist < TYPEHASHSIZE; typep = typenextp)  {
                                     /* run through all the HOSTs */
    if (typep -> name == NULL) {      /* then finished this list */
      typenextp = typehead[++onlist];  /* so look at the next list */
    }
    else {
      if ((typesortby == BYBYTES &&
           (typep -> bytes / (total_bytes / 10000)) < typefloor) ||
          (typesortby != BYBYTES && typep -> reqs < typefloor)) {
        typenextp = typep -> next;
      }
    	else {
        tnumber++;
	no_typesort++;
        type_max_reqs = MAX(typep -> reqs, type_max_reqs);
        type_max_bytes = MAX(typep -> bytes, type_max_bytes);
        if ((typesorthead -> name == NULL) ||
            (typesortby == BYBYTES && typep -> bytes > typesorthead -> bytes) ||
            (typesortby == BYREQUESTS &&
             typep -> reqs > typesorthead -> reqs) ) {
          /* if it's before the first item currently on the list, slot it in */
          typenextp = typep -> next;   /* the next one we're going to look at */
          typep -> next = typesorthead;
          typesorthead = typep;
        }
        else {              /* otherwise compare with the ones so far */
          finished = OFF;
          typelastp = typesorthead;
          for (typep2 = typesorthead -> next; typep2 -> name != NULL
               && (!finished) ;
               typep2 = typep2 -> next) {
            if ((typesortby == BYBYTES && typep -> bytes > typep2 -> bytes) ||
                (typesortby == BYREQUESTS && typep -> reqs > typep2 -> reqs) ) {
          /* if typep comes before typep2 in the chosen ordering, slot it in */
              typenextp = typep -> next;
              typep -> next = typep2;
              typelastp -> next = typep;
              finished = ON;
            }
            typelastp = typep2;
          }
          if (!finished) {         /* we've reached the end of the list; */
                                   /* slot it in at the end */
            typenextp = typep -> next;
            typep -> next = typep2;
            typep2 -> name = NULL;
            typelastp -> next = typep;
          }
        }
        }
    }
    
    typep = typenextp;   /* so, on to the next one */

  }        /* end for running through all filenames */

  return(typesorthead);

}    /* end typesort */



/*** Again, the domain report is a bit different because the
   structure is stored differently ***/

int domsort(void)
{
  extern int onumber;
  extern struct domain *domainhead[];
  extern int domsortby;
  extern int domfloor;
  extern double total_bytes;
  extern int dom_max_reqs;
  extern double dom_max_bytes;
  extern int no_domsort;

  int i, j, k;
  int firstdom;  /* the numerical index of the first domain; we return this */
  int domnextj;
  struct domain *domp, *domlastp;
  flag finished;

  onumber = 0;
  firstdom = DOMHASHSIZE - 2; /* start with unknown domains at front of list */
  dom_max_reqs = domainhead[firstdom] -> reqs;
  dom_max_bytes = domainhead[firstdom] -> bytes;
  domainhead[firstdom] -> nexti = -1;
  j = DOMHASHSIZE - 1; /* the domain we are on; start with numerical domains */
  while (j >= 0) {              /* run through all the domains */
    domp = domainhead[j];
    domnextj = domp -> nexti;
                            /* the one we're going to look at after this one */
    if (!((domsortby == BYREQUESTS && domp -> reqs == 0) ||
	  (domsortby == BYBYTES &&
	   (domp -> bytes / (total_bytes / 10000)) < domfloor) ||
	  (domsortby == BYREQUESTS && domp -> reqs < domfloor))) {
                                   /* else we don't want it */
      onumber++;
      no_domsort++;
      dom_max_reqs = MAX(domp -> reqs, dom_max_reqs);
      dom_max_bytes = MAX(domp -> bytes, dom_max_bytes);
      if ((domsortby == BYBYTES &&
	   domp -> bytes > domainhead[firstdom] -> bytes) ||
	  (domsortby == BYREQUESTS &&
	   domp -> reqs > domainhead[firstdom] -> reqs) ){
	/* if it's before the first item currently on the list, slot it in */
	domp -> nexti = firstdom;
	firstdom = j;
      }
      else {        /* otherwise compare with the ones so far */
	finished = OFF;
	domlastp = domainhead[firstdom];
	if (domfloor < 0)
	  k = domfloor;
	else
	  k = 1;
	for (i = domainhead[firstdom] -> nexti;
	     i >= 0 && (!finished) && (k++) != -1;
	     i = domainhead[i] -> nexti) {
	  if ((domsortby == BYBYTES &&
	       domp -> bytes > domainhead[i] -> bytes) ||
	      (domsortby == BYREQUESTS &&
	       domp -> reqs > domainhead[i] -> reqs) ){
	    /* if domp comes before domp2 in the chosen ordering, slot it in */
	    domp -> nexti = i;
	    domlastp -> nexti = j;
	    finished = ON;
	  }
	  domlastp = domainhead[i];
	}
	if (!finished) {
	  domp -> nexti = -1;   /* meaning, last item on the list */
	  domlastp -> nexti = j;
	}
      }
    }
    
    j = domnextj;   /* so, on to the next one */
	
  }        /* end while j >= 0 */

  return(firstdom);

}   /* end domsort */
