/*** prostat  1.32***/

/*** prostat.c; the main function ***/

#include "analhead2.h"

int main(int argc, char **argv)
{
  extern struct tm *localtime();
  extern int is_it_a_hit();
  extern flag wantfile();                   /* in alias.c */
  extern flag wanthost();                   /* in alias.c */
  extern void reqhashadd();                 /* in hash.c */
  extern void domhashadd();                 /* in hash.c */
  extern void protohashadd();                 /* in hash.c */
  extern void hosthashadd();                /* in hash.c */
  extern void neighbourhashadd();                /* in hash.c */
  extern void typehashadd();                /* in hash.c */
  extern void mimehashadd();                /* in hash.c */
  extern void approxhosthashadd();          /* in hash.c */
  extern struct url *reqsort();             /* in hash.c */
  extern struct proto *protosort();             /* in hash.c */
  extern struct request *requestsort();             /* in hash.c */
  extern struct host *hostsort();           /* in hash.c */
  extern struct file_types *typesort();           /* in hash.c */
  extern struct mime_types *mimesort();           /* in hash.c */
  extern int domsort();                     /* in hash.c */
  extern void addmonthlydata();             /* in hash.c */
  extern void adddailydata();               /* in hash.c */
  extern void addweeklydata();              /* in hash.c */
  extern void initialise();                 /* in init.c */
  extern void printvbles();                 /* in init.c */
  extern void output();                     /* in output.c */
  extern int sscan_line();               /* in sscanf.c */
  extern int sscanf_ncsaold();              /* in sscanf.c */
  extern int strtomonth();                  /* in utilities.c */
  extern int dayofdate();                   /* in utilities.c */
  extern int minsbetween();                 /* in utilities.c */
  extern long timecode();                   /* in utilities.c */
  extern struct timestruct startofweek();   /* in utilities.c */
  extern void *xmalloc();                   /* in utilities.c */
  extern void requesthashadd();

  extern char commandname[];    /* all global vars declared in init.c */
  extern char logfile[];
  extern FILE *lf;
  extern flag vblesonly, maskq;
  extern flag mq, hq, dq, Sq, oq, Pq, sq, Rq, tq;
  extern int domfloor, sitesfloor, protofloor, hostfloor;
  extern struct timestruct firsttime, lasttime, fromtime, totime;
  extern int total_succ_reqs;
  extern int total_reqs;
  extern int req_hit;
  extern double req_hit_size;
  extern int tcp_reqs, err_reqs;
  extern double tcp_size, err_size;
  extern int udp_reqs;
  extern int req_udp_hit_total;
  extern int req_udp_miss_total;
  extern int req_tcp_hit_total;
  extern int req_tcp_miss_total;
  extern double req_udp_hit_total_size;
  extern double req_udp_miss_total_size;
  extern double req_tcp_hit_total_size;
  extern double req_tcp_miss_total_size;
  extern double req_udp_hit_total_time;
  extern double req_udp_miss_total_time;
  extern double req_tcp_hit_total_time;
  extern double req_tcp_miss_total_time;
  extern double udp_size;
  extern int corrupt_lines, other_lines;
  extern double total_bytes;
  extern int dailyreq[], hourlyreq[];
  extern void double3printf();
  extern double dailysize[], hourlysize[];
  extern int rnumber, inumber, onumber, Snumber, tnumber, mnumber;
  extern flag debug;
  extern char firstday[];
  extern char outfile[];

  char inputline[MAXLINELENGTH];  /* a particular input line */
  int linetype;          /* COMMON, NCSAOLD or CORRUPT */
  char hostn[MAXSTRINGLENGTH];
  flag is_an_alias;
  char method[15];
  char protocole[MAXSTRINGLENGTH];
  int date, year, hr, min, monthno, req_tcp, elapse_time;
  long thistimecode;
  char month[4];
  int day;
  char rhost[MAXSTRINGLENGTH], tag2[MAXSTRINGLENGTH], neighbour[MAXSTRINGLENGTH];
  char filetype[10];
  int code;
  double bytes;    /* long is not big enough; double has more sig. figs,
		      and copes with overflow automatically. */
  char bytestr[16];
  char tag[MAXSTRINGLENGTH];
  char file[MAXSTRINGLENGTH];
  FILE *outf;

  struct url *urlsorthead;
  struct proto *protosorthead;
  struct request *requestsorthead;
  struct host *hostsorthead;
  struct file_types *typesorthead;
  struct mime_types *mimesorthead;
  int firstdom;             /* for sorting */
  int numline, lastnum;
 

  flag wantthisone = TRUE;  /* whether we want to analyse a particular entry */


  int rc, i;
  char s[20];

  /*** Initialisation ***/

  initialise(argc, argv);

  if (vblesonly)
    printvbles();    /* which also exits */

  /*** Now start scanning ***/

  if (strcmp(logfile, "stdin") == 0)   /* first check the logfile exists */
    lf = stdin;
  else {
    lf = fopen(logfile, "r");
    if (lf == NULL) {
      fprintf(stderr, "%s: Error: Failed to open log file %s: exiting\n",
	      commandname, logfile);
      exit(ERR);
    }
  }

  numline=0;
  lastnum=0;
  fprintf(stdout, "Number of lines computed :\t");
  while(fgets(inputline, MAXLINELENGTH, lf) != NULL) {
    numline++;
    if(numline - lastnum == 100) {
        if(numline > 100) {
	  sprintf(s, "%d", numline-100);
	  for(i=0; i< strlen(s); i++) {
 	    fprintf(stdout, "\b");
 	  } 
	}
        fprintf(stdout, "%d", numline);
        lastnum = numline;
    }
    if(debug)
	fprintf(stderr, "===================================\n%d : %s\n", numline, inputline);
    linetype = CORRUPT;   /* paranoia :) */
    rc = sscan_line(inputline, hostn, &is_an_alias, &date, month, &year, &hr, &min,
		rhost, &code, tag, bytestr, protocole, method, filetype, tag2, neighbour, &elapse_time);
    if(rc == 0)
	linetype = COMMON;

    monthno = strtomonth(month);

    if (linetype != CORRUPT && strcmp(protocole, "cache_object") != 0 && strcmp(protocole, "disk_object") != 0 && strcmp(protocole, "cache_config") != 0) {

      monthno = strtomonth(month);
      thistimecode = timecode(date, monthno, year, hr, min);

      if (maskq) {   /* then check if we want this entry */
	wantthisone = (thistimecode > fromtime.code &&
		       thistimecode < totime.code && wantfile(rhost) &&
		       wanthost(hostn));
      }

      if(strcmp(firstday, "\0") == 0 ) {
		sprintf(firstday, "%d.%02d.%02d", year, monthno+1, date);

/* fprintf(stderr, "firstday = %s\n", firstday); */
		if (strcmp(outfile, "stdout") != 0) {
			sprintf(file, "%s/stat%s.html", outfile, firstday);
			if ((outf = fopen(file, "w")) == NULL) {
				fprintf(stderr,
       				"%s: Error: failed to open output file %s/stat%s.html for writing.\n",
       				commandname, outfile, firstday);
				exit(ERR);   /* test now so we don't do the processing THEN find out */
			}
			else
				fclose(outf);  /* no need to hold it open */
  		}
      }


      if (!wantthisone || strcmp(tag, "TCP_DENIED") == 0 || strcmp(tag, "UDP_DENIED") == 0 || strcmp(tag, "UDP_INVALID") == 0 || strcmp(tag, "ERR_INVALID_URL") == 0) {
	++other_lines;
	if (debug)
                fprintf(stderr, "N: %s", inputline);
	}
      else {
        ++total_reqs;

	  bytes = atof(bytestr);
	  if (is_it_a_hit(tag) == 1) {
                req_hit++;
                req_hit_size += bytes;
if (debug)
                fprintf(stderr, "Hit : %s\n", tag);
          }
	  else {
if (debug)
                fprintf(stderr, "No Hit : %s\n", tag);
}
          req_tcp=0;
	  if(strncmp(tag, "UDP", 3) == 0) {
		udp_reqs++;
		udp_size  += bytes;
		if(strstr(tag, "HIT") != NULL) {
			req_udp_hit_total++;
			req_udp_hit_total_size += bytes;
			req_udp_hit_total_time += elapse_time;
		}
		else {
			req_udp_miss_total++;
			req_udp_miss_total_size += bytes;
			req_udp_miss_total_time += elapse_time;
		}
	  }
	  else {
	  if(strncmp(tag, "TCP", 3) == 0) {
		tcp_reqs++;
		tcp_size += bytes;
		if(strstr(tag, "HIT") != NULL) {
			req_tcp_hit_total++;
			req_tcp_hit_total_size += bytes;
			req_tcp_hit_total_time += elapse_time;
		}
		else {
			req_tcp_miss_total++;
                        req_tcp_miss_total_size += bytes;
                        req_tcp_miss_total_time += elapse_time;
		}
		req_tcp=1;
	  }
	  else {
	  	if(strncmp(tag, "ERR", 3) == 0) {
			err_reqs++;
			err_size += bytes;
	  	}
	  }
	  }

	  day = dayofdate(date, monthno, year);
	  
	  ++total_succ_reqs;
	  total_bytes += bytes;
	  if(debug)
                fprintf(stderr, "total_reqs = %d, total_succ_reqs = %d, req_hit = %d\n", total_reqs, total_succ_reqs, req_hit);

	  if (total_succ_reqs == 1) {  /* i.e. this is the first */
	    firsttime.date = date;
	    firsttime.monthno = monthno;
	    firsttime.year = year;
	    firsttime.hr = hr;
	    firsttime.min = min;
	    firsttime.code = thistimecode;
	    lasttime.date = date;
	    lasttime.monthno = monthno;
	    lasttime.year = year;
	    lasttime.hr = hr;
	    lasttime.min = min;
	    lasttime.code = thistimecode;
	  }
	  
	  /* date cataloguing */


	  ++dailyreq[day];  /* these are so little work, it's cheaper just to */
	  ++hourlyreq[hr];
                            /* do them rather than to check dq and hq */
	  dailysize[day] += bytes;
	  hourlysize[hr] += bytes;

	  if (thistimecode < firsttime.code) {
	    firsttime.date = date;
	    firsttime.monthno = monthno;
	    firsttime.year = year;
	    firsttime.hr = hr;
	    firsttime.min = min;
	    firsttime.code = thistimecode;
	  }

	  if (thistimecode > lasttime.code) {
	    lasttime.date = date;
	    lasttime.monthno = monthno;
	    lasttime.year = year;
	    lasttime.hr = hr;
	    lasttime.min = min;
	    lasttime.code = thistimecode;
	  }

	  /* Now for the request report. We ignore all sitename conversions
	     and aliases until outside the loop; it is more efficient to do
	     them just once at the end.  */
	  
	  if (sq || Pq) {
	    if(is_it_a_hit(tag) == 1)
	    	reqhashadd(rhost, 1, bytes, 1, OFF);
	    else
		reqhashadd(rhost, 1, bytes, 0, OFF);
	  }
	  
	  /* Now for the hostname count. Just the same as above. This time,
	     however, we don't do one if the domain report is on and this
	     is off, because this takes up a lot of memory. */
	  if(Pq)
		protohashadd(protocole, 1, bytes) ;
	  if(Rq)
		requesthashadd(tag, 1, bytes, elapse_time) ;

	  if(Sq == ON) 
		if(strncmp(tag2, "SIBLING", 7) == 0 || (strcmp(tag2, "NEIGHBOR_HIT") == 0 && strcmp(neighbour, "-") != 0)) {
			neighbourhashadd(neighbour, 1, bytes);
		}

if(debug)
	fprintf(stderr, "req_tcp = %d\n", req_tcp);
		if(is_it_a_hit(tag) == 1) 
	      		hosthashadd(hostn, 1, bytes, elapse_time, 1, req_tcp, is_an_alias);
		else
	      		hosthashadd(hostn, 1, bytes, 0, 0, req_tcp, is_an_alias);

	  if ((tq == ON || mq == ON) && strcmp(filetype, "\0") != 0) {
		typehashadd(filetype, "", 1, bytes);
	  }

	  /* If there is no exact hostname count, do the domain report now */

	  if (oq) {
	    if(is_it_a_hit(tag) == 1)
	    	domhashadd(rhost, 1, bytes, 1);
	    else
		domhashadd(rhost, 1, bytes, 0);
	  }

      }  /* end if want this one */
	
    }   /* end if linetype != CORRUPT */

    else {   /* line is corrupt */
	if(linetype == CORRUPT) {
      		++corrupt_lines;
		if(debug)
			fprintf(stderr, "\nC: %s\nrc : %d\n", inputline, rc);
	}
     	if (strchr(inputline, '\n') == NULL) {
                      /* line corrupt by being too long; */
		fscanf(lf, "%*[^\n]");      /* read to end of line */
		if (debug != 0)
	  		fprintf(stderr, "\n");
	}
    }
  
  }   /*** end of main loop (reading logfile) ***/

  fclose(lf);
  sprintf(s, "%d", numline-100);
  for(i=0; i< strlen(s); i++) {
 	    fprintf(stdout, "\b");
  } 
  fprintf(stdout, "%d\n", numline);
  fprintf(stderr, "Begin : %s\n", firstday);
  fprintf(stderr, "End   : %d.%02d.%02d\n", year, monthno+1, date);

  if (total_succ_reqs == 0) {
    mq = OFF;
    dq = OFF;
    hq = OFF;
    oq = OFF;
    Pq = OFF;
    sq = OFF;
    Rq = OFF;
    Sq = OFF;
    tq = OFF;
  }

  else {   /* there are things to report */

    /* now for the checking and sorting */
    
    if (sq) {
      urlsorthead = reqsort();
      if (rnumber == 0)
	sq = OFF;
      else if (rnumber < -sitesfloor)
                             /* i.e. we want top n but there are only m < n */
	sitesfloor = 0;
    }

    if (Pq) {
      protosorthead = protosort();
      if (inumber == 0)
	Pq = OFF;
      else if (inumber < -protofloor)
	protofloor = 0;
    }

    if (Rq) {
      requestsorthead = requestsort();
      if (inumber == 0)
        Rq = OFF;
    }

    if (Sq) {
      hostsorthead = hostsort();
      if (Snumber == 0)
	Sq = OFF;
      else if (Snumber < -hostfloor)
	hostfloor = 0;
    }

    if (tq) {
      typesorthead = typesort();
      if (tnumber == 0)
        tq = OFF;
    }

    if (mq) {
      mimesorthead = mimesort();
      if (mnumber == 0)
        mq = OFF;
    }


    if (oq) {
      firstdom = domsort();
      if (onumber == 0)
	oq = OFF;
      else if (onumber < -domfloor)
	domfloor = 1;
    }

  }    /* end else (there are things to report */

  /*** Finally, do all the output ***/

  output(urlsorthead, protosorthead, requestsorthead, hostsorthead, typesorthead, mimesorthead, firstdom);

  return(OK);
   
}
